%%
%% %CopyrightBegin%
%%
%% Copyright Ericsson AB 2012. All Rights Reserved.
%%
%% The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved online at http://www.erlang.org/.
%%
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%%
%% %CopyrightEnd%

%% This is a TCP client used for network benchmarking.
%% The client sets up connections and writes a fixed amount of data
%% that is expected to be echoed back by the server.

-module(netmarks_client).

-compile(export_all).

%-define(DEBUG,true).
-ifdef(DEBUG).
-define(DBG_TRACE(F,L),io:format(F,L)).
-else.
-define(DBG_TRACE(F,L),void).
-endif.


-define(DEFAULT_ITERATIONS,1000000).
-define(DEFAULT_HEADER_SZ,0).
-define(DEFAULT_PACKET_SZ,1000).
-define(DEFAULT_PORT,4999).
-define(DEFAULT_PARALLEL,1).
-define(DEFAULT_HOST,"localhost").
-define(DEFAULT_SYNC,false).
-define(DEFAULT_ACTIVE,once).


run() ->
    run([]).

help() ->
    def_opts().

def_opts() -> % order is significant
    [{parallel,?DEFAULT_PARALLEL},
     {iterations,?DEFAULT_ITERATIONS},
     {host,?DEFAULT_HOST},
     {port,?DEFAULT_PORT},
     {packet_size,?DEFAULT_PACKET_SZ},
     {header_size,?DEFAULT_HEADER_SZ},
     {sync,?DEFAULT_SYNC},
     {active,?DEFAULT_ACTIVE}].

run(Opts) when is_list(Opts) ->
    DefOpts = def_opts(),
    verify_opts(Opts, DefOpts),
    [Parallel | RunArgs] = [proplists:get_value(K,Opts,V) || {K,V} <- DefOpts],
    run(Parallel, RunArgs, []).

run(0, _, Pids) ->
    TotBytes = join(Pids, 0),
    io:format("Done. Sent and received a total of ~p bytes\n", [TotBytes]);
run(Parallel, Opts, Pids) ->
    Pid = spawn_link(?MODULE,client_start,[self(), Opts]),
    run(Parallel-1, Opts, [Pid | Pids]).

client_start(Papa, [Iters, Host, Port, Psz, Hsz, Sync, Active]) ->
    BuffSize = 1460*2,
    {ok, Sock} =
	gen_tcp:connect(Host, Port, [{active,Active}, {packet,Hsz}, binary,
				     {reuseaddr, true},{recbuf,BuffSize},
				     {sndbuf,BuffSize}]),
    Bytes = case Sync of
		false ->
		    spawn_link(?MODULE, writer_start,
			       [self(), Sock, Hsz, build_packet(Psz), Iters]),
		    read_loop(Sock, Hsz, 0, undefined, Active);

		true ->
		    sync_loop(Sock, Hsz, build_packet(Psz), Iters, 0, Active)
    end,
    Papa ! {self(), done, Bytes}.


sync_loop(_, _, _, 0, TotBytes, _) ->
    TotBytes;
sync_loop(Sock, Hsz, Packet, Iters, TotBytes0, Active) ->
    gen_tcp:send(Sock, Packet),
    Psz = byte_size(Packet),
    ?DBG_TRACE("~p: Sent ~p bytes\n", [self(), Psz]),
    sync_reader(Sock, Psz, Active),
    sync_loop(Sock, Hsz, Packet, Iters-1, TotBytes0 + Psz + Hsz, Active).

sync_reader(_Sock, 0, _) -> ok;
sync_reader(Sock, BytesLeft, false) ->
    case gen_tcp:recv(Sock,0) of
	{ok, Data} ->
	    Psz = byte_size(Data),
	    sync_reader(Sock, BytesLeft - Psz, false)
    end;
sync_reader(Sock, BytesLeft, Active) ->
    receive
	{tcp, Sock, Data} ->
	    Psz = byte_size(Data),
	    ?DBG_TRACE("~p: Read ~p of ~p bytes\n", [self(), Psz, BytesLeft]),
	    if Active == once -> inet:setopts(Sock,[{active,once}]);
	       true -> ok end,
	    sync_reader(Sock, BytesLeft - Psz, Active);

	Other ->
	    io:format("Reader ~p got unexpected message: ~p\n", [self(), Other]),
	    exit("unexpected msg")
    end.


writer_start(Reader, Sock, Hsz, Packet, Iters) ->
    Written = write_loop(Sock, Hsz, Packet, Iters, 0),
    ?DBG_TRACE("Writer ~p done writing ~p bytes\n", [self(), Written]),
    Reader ! {done_writing, Written}.

write_loop(_, _, _, 0, Written) ->
    Written;
write_loop(Sock, Hsz, Packet, N, Written) ->
    gen_tcp:send(Sock, Packet),
    write_loop(Sock, Hsz, Packet, N-1, Written+byte_size(Packet)+Hsz).

read_loop(_, _, ReadAll, ReadAll, _Active) ->
    ReadAll;
read_loop(Sock, Hsz, ReadBytes0, TotBytes, false) ->
    case gen_tcp:recv(Sock, 0, 200) of
	{ok, Data} ->
	    ReadBytes1 = ReadBytes0 + byte_size(Data) + Hsz,
	    ?MODULE:read_loop(Sock, Hsz, ReadBytes1, TotBytes, false);
	{error, timeout} ->
	    receive
		{done_writing, Written} when Written >= ReadBytes0 ->
		    ?DBG_TRACE("Reader ~p read ~p bytes of ~p written bytes.\n",
			       [self(), ReadBytes0, Written]),
		    undefined = TotBytes,
		    ?MODULE:read_loop(Sock, Hsz, ReadBytes0, Written, false);
		Other ->
		    io:format("Reader ~p got unexpected message: ~p\n", [self(), Other]),
		    exit("unexpected msg")
	    after 0 ->
		    ?MODULE:read_loop(Sock, Hsz, ReadBytes0, TotBytes, false)
	    end
    end;
read_loop(Sock, Hsz, ReadBytes0, TotBytes, Active) ->
    receive
        {tcp, Sock, Data} ->
	    ReadBytes1 = ReadBytes0 + byte_size(Data) + Hsz,
	    ?DBG_TRACE("Reader ~p got ~p bytes, bytes left: ~p\n",
		       [self(), byte_size(Data), bytes_left(ReadBytes1,TotBytes)]),
	    if Active == once -> inet:setopts(Sock,[{active,once}]);
	       true -> ok end,
            ?MODULE:read_loop(Sock, Hsz, ReadBytes1, TotBytes, Active);

	{done_writing, Written} when Written >= ReadBytes0 ->
	    ?DBG_TRACE("Reader ~p read ~p bytes of ~p written bytes.\n",
		       [self(), ReadBytes0, Written]),
	    undefined = TotBytes,
	    ?MODULE:read_loop(Sock, Hsz, ReadBytes0, Written, Active);

	Other ->
	    io:format("Reader ~p got unexpected message: ~p\n", [self(), Other]),
	    exit("unexpected msg")

    after 1000 ->
	    ?DBG_TRACE("Reader ~p still waiting for ~p bytes\n",
		       [self(), bytes_left(ReadBytes0, TotBytes)]),
	    ?MODULE:read_loop(Sock, Hsz, ReadBytes0, TotBytes, Active)
    end.

bytes_left(_, undefined) -> undefined;
bytes_left(Read, Total) -> Total - Read.

build_packet(Sz) ->
    << <<B>> || B <- packet_bytes(Sz)>>.

packet_bytes(Sz) ->
    packet_bytes($a, Sz, []).

packet_bytes(_, 0, Acc) ->
    lists:reverse(Acc);
packet_bytes($z, Left, Acc) ->
    packet_bytes($a, Left-1, [$z | Acc]);
packet_bytes(Char, Left, Acc) ->
    packet_bytes(Char+1, Left-1, [Char | Acc]).


join([], TotBytes) ->
    TotBytes;
join([Pid | Tail], TotBytes) ->
    receive
	{Pid, done, Bytes} -> join(Tail, TotBytes + Bytes)
    end.

verify_opts([], _) ->
    ok;
verify_opts([{Key,_} | Tail], DefOpts) ->
    case proplists:is_defined(Key, DefOpts) of
	true -> ok;
	false ->
	    Reason = lists:flatten(io_lib:format("Unknown option '~p'", [Key])),
	    exit(Reason)
    end,
    verify_opts(Tail, DefOpts).
